# typed: ignore

# Copyright (c) 2015 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.com/terms.html

require 'net/http'
require 'uri'

require 'sqreen/kit/loggable'
require 'sqreen/kit/retry_policy'
require 'sqreen/kit/http_client/authentication_error'
require 'sqreen/kit/http_client/unexpected_status_error'

module Sqreen
  module Kit
    # An http client doing JSON requests
    class HttpClient
      include Loggable

      DEFAULT_CONNECT_TIMEOUT_S = 5
      DEFAULT_READ_TIMEOUT_S = 30

      attr_reader :http

      # @return [URI]
      attr_reader :base_url

      # @param server_url [String]
      # @param retry_policy [Sqreen::Kit::RetryPolicy]
      def initialize(server_url, retry_policy, opts = {})
        @base_url = parse_uri(server_url)

        @retry_policy = retry_policy

        @http = ::Net::HTTP.new(@base_url.host, @base_url.port,
                                opts[:proxy_address] || :ENV,
                                opts[:proxy_port], opts[:proxy_user],
                                opts[:proxy_pass])
        @http.use_ssl = @base_url.scheme.downcase == 'https'
        @http.cert_store = opts[:certificate_store]
        @http.verify_mode = OpenSSL::SSL::VERIFY_NONE if ENV['SQREEN_SSL_NO_VERIFY'] # for testing

        @http.open_timeout = opts[:connect_timeout] || DEFAULT_CONNECT_TIMEOUT_S
        @http.ssl_timeout = opts[:read_timeout] || DEFAULT_READ_TIMEOUT_S
        @http.read_timeout = opts[:read_timeout] || DEFAULT_READ_TIMEOUT_S

        @req_nb = 1

        @static_headers = init_static_headers
      end

      # @param path [String]
      # @param data [String]
      # @param headers [Hash{String=>String}]
      def post(path, data, headers = {})
        request(:POST, path, data, headers)
      end

      def get(path, headers = {})
        request(:GET, path, nil, headers)
      end

      private

      # @return [URI::HTTP]
      def parse_uri(uri)
        # This regexp is the Ruby constant URI::PATTERN::HOSTNAME augmented
        # with the _ character that is frequent in Docker linked containers.
        re = '(?:(?:[a-zA-Z\\d](?:[-_a-zA-Z\\d]*[a-zA-Z\\d])?)\\.)*(?:[a-zA-Z](?:[-_a-zA-Z\\d]*[a-zA-Z\\d])?)\\.?'
        parser = URI::Parser.new HOSTNAME: re
        uri += '/' unless uri.end_with? '/'
        parser.parse(uri)
      end

      def with_retry(&block)
        @retry_policy.execute(&block)
      end

      def connect
        return if connected?
        logger.warn { "Connecting to #{base_url}..." }

        begin
          with_retry do
            http.start
          end
        rescue StandardError => e
          logger.warn "Cannot connect to #{base_url}: #{e.message}"
          raise
        else
          logger.warn 'Connection success'
        end
      end

      def connected?
        http.started?
      end

      def disconnect
        http.finish if connected?
      end

      def init_static_headers
        platform = RUBY_PLATFORM
        platform += " #{ENV_JAVA['java.runtime.version']}" if defined?(ENV_JAVA)
        ua_string = "sqreen-kit/#{Sqreen::Kit.version} " \
                    "(#{platform}) ruby/#{RUBY_VERSION}"
        ua_string += " jruby/#{JRUBY_VERSION}" if defined?(JRUBY_VERSION)
        { 'User-Agent' => ua_string }
      end

      def request(method, path, data, headers = {})
        connect

        now = Time.now

        headers = @static_headers.merge(headers)
        @req_nb += 1

        path = (base_url + path).path

        with_retry do
          logger.debug do
            format('%s %s %s (%s)',
                   method, path, data.inspect, headers.inspect)
          end

          resp = case method
                 when :GET
                   http.get(path, headers)
                 when :POST
                   http.post(path, data, headers)
                 end

          result = process_response resp

          logger.debug do
            format('%s %s (DONE: %s in %f ms)',
                   method, path, resp && resp.code, (Time.now - now) * 1000)
          end

          result
        end # end with_retry
      end

      # @param [Net::HTTPResponse] resp
      def process_response(resp)
        body = resp.body
        logger.debug { "Body of response: #{body}" }

        if resp.code == '401' || resp.code == '403'
          raise AuthenticationError.new(resp.code, body)
        end

        unless %w[200 201 202].include?(resp.code)
          raise UnexpectedStatusError.new(resp.code, body)
        end

        {
          code: resp.code,
          body: body,
        }
      end
    end
  end
end
