# typed: ignore

# Copyright (c) 2015 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.com/terms.html

require 'json'

module Sqreen
  module Kit
    module Signals
      # Provides a helper constructor, default to_h, and mandatory field checking
      module DtoHelper
        DO_NOT_CONVERT_TYPES = [
          NilClass, Hash, Array, Numeric, TrueClass, FalseClass
        ].freeze

        RFC_3339_FMT = '%Y-%m-%dT%H:%M:%S.%3N%:z'.freeze

        module ClassMethods
          # method should have been defined initially in an ancestor
          def validate_str_attr(attr, regex)
            define_method(:"#{attr}=") do |val|
              unless val =~ regex
                raise "Unexpected format for attribute #{attr}: " \
                      "'#{val}' does not match #{regex}"
              end
              super(val)
            end
          end

          # All the classes/modules in the ancestor chain
          # including DtoHelper
          def all_implementing_modules
            ancestors
              .select { |c| c != DtoHelper && c.ancestors.include?(DtoHelper) }
          end

          def attributes_for_to_h
            @all_attributes ||= begin
              all_implementing_modules
                .map(&:attributes_for_to_h_self)
                .reduce(:+)
                .uniq
            end
          end

          def attributes_for_to_h_self
            methods = public_instance_methods(false)

            methods.reject { |m| m.to_s.end_with? '=' }
                   .select { |m| methods.include?(:"#{m}=") }
          end

          def add_mandatory_attrs(*args)
            @self_mandatory_attrs ||= []
            @self_mandatory_attrs += args
          end

          def readonly_attrs(*attrs)
            attrs.each do |attr|
              define_method :"#{attr}=" do |value|
                raise "Attribute #{attr} is read-only" unless public_send(attr).nil?
                super(value)
              end
            end
          end

          def self_mandatory_attrs
            @self_mandatory_attrs ||= []
          end

          def mandatory_attrs
            @mandatory_attrs ||= begin
              all_implementing_modules
                .map(&:self_mandatory_attrs)
                .reduce(:+)
                .uniq
            end
          end

          def attr_accessor_time(attr)
            define_method :"#{attr}=" do |value|
              unless value.is_a?(Time)
                unless value.is_a?(String)
                  raise ArgumentError, 'expected Time or String object'
                end
                unless value =~ /\A\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}/
                  raise ArgumentError, "Invalid time format for #{value}"
                end
              end
              instance_variable_set("@#{attr}", value)
            end

            define_method attr do
              var = "@#{attr}"
              return nil unless instance_variable_defined?(var)

              cur_val = instance_variable_get(var)

              return nil if cur_val.nil?
              return cur_val if cur_val.is_a?(String)
              cur_val.strftime(RFC_3339_FMT)
            end
          end

          def included(mod)
            # make sure that classes/modules indirectly including DtoHelper
            # also have their singleton class including ClassMethods
            return if mod.singleton_class.ancestors.include?(ClassMethods)
            mod.extend(ClassMethods)
          end
        end

        def self.included(mod)
          mod.extend(ClassMethods)
        end

        def initialize(values = {})
          values.each do |attr, val|
            public_send("#{attr}=", val)
          end
        end

        def compact_hash(h)
          h.delete_if { |_k, v| v.nil? }
        end

        def to_h
          check_mandatories

          res = {}
          self.class.attributes_for_to_h.each do |attr|
            value = public_send(attr)
            if (value.class.ancestors & DO_NOT_CONVERT_TYPES).empty? && \
               value.respond_to?(:to_h)
              value = value.to_h
            end

            res[attr] = value unless value.nil?
          end
          res
        end

        def append_to_h_filter(proc)
          @filters ||= []
          @filters << proc
        end

        def to_json
          return to_h.to_json unless instance_variable_defined?(:@filters)

          res = @filters.reduce(to_h) { |accum, filter| filter[accum] }
          res.to_json
        end

        private

        def check_mandatories
          self.class.mandatory_attrs.each do |attr|
            if public_send(attr).nil?
              raise "The attribute #{attr} is not set in #{inspect}"
            end
          end
        end
      end
    end
  end
end
