# typed: ignore

# Copyright (c) 2015 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.com/terms.html

# Includes code from Airbrake Technologies, Inc., under the MIT licence
# see https://github.com/airbrake/airbrake-ruby/blob/c172657dcd4eeaf62ea23f0f7d6111be2373ea59/LICENSE.md

require 'sqreen/kit/loggable'

module Sqreen
  module Kit
    module StackTrace
    end
  end
end

class << Sqreen::Kit::StackTrace
  module Patterns
    # @return [Regexp] the pattern that matches standard Ruby stack frames,
    #   such as ./spec/notice_spec.rb:43:in `block (3 levels) in <top (required)>'
    RUBY = %r{\A
            (?<file>.+)       # Matches './spec/notice_spec.rb'
            :
            (?<line>\d+)      # Matches '43'
            :in\s
            `(?<function>.*)' # Matches "`block (3 levels) in <top (required)>'"
            \z}x

    # @return [Regexp] the pattern that matches JRuby Java stack frames, such
    #  as org.jruby.ast.NewlineNode.interpret(NewlineNode.java:105)
    JAVA = %r{\A
            (?<function>.+)  # Matches 'org.jruby.ast.NewlineNode.interpret'
            \(
            (?<file>
              (?:uri:classloader:/.+(?=:)) # Matches '/META-INF/jruby.home/protocol.rb'
              |
              (?:uri_3a_classloader_3a_.+(?=:)) # Matches 'uri_3a_classloader_3a_/gems/...'
              |
              [^:]+        # Matches 'NewlineNode.java'
            )
            :?
            (?<line>\d+)?  # Matches '105'
            \)
            \z}x

    # @return [Regexp] the pattern that tries to assume what a generic stack
    #   frame might look like, when exception's backtrace is set manually.
    GENERIC = %r{\A
              (?:from\s)?
              (?<file>.+)              # Matches '/foo/bar/baz.ext'
              :
              (?<line>\d+)?            # Matches '43' or nothing
              (?:
                in\s`(?<function>.+)'  # Matches "in `func'"
                |
                :in\s(?<function>.+)   # Matches ":in func"
              )?                       # ... or nothing
              \z}x
  end

  include Sqreen::Kit::Loggable

  # Parses an exception's backtrace.
  #
  # @param exception [Exception] The exception, which contains a backtrace to
  #   parse
  # @return [Array<Hash{Symbol=>String,Integer}>] the parsed backtrace
  def parse(exception)
    return [] if exception.backtrace.nil? || exception.backtrace.none?
    do_parse(exception)
  end

  def parse_backtrace(backtrace)
    # assume normal ruby backtrace
    backtrace.map do |stackframe|
      stack_frame(Patterns::RUBY, stackframe)
    end
  end

  private

  def do_parse(exception)
    regexp = best_regexp_for(exception)

    exception.backtrace.map do |stackframe|
      stack_frame(regexp, stackframe)
    end
  end

  def best_regexp_for(exception)
    if java_exception?(exception)
      Patterns::JAVA
    else
      Patterns::RUBY
    end
  end

  # Checks whether the given exception was generated by JRuby's VM.
  #
  # @param [Exception] exception
  # @return [Boolean]
  def java_exception?(exception)
    if defined?(Java::JavaLang::Throwable) &&
       exception.is_a?(Java::JavaLang::Throwable)
      return true
    end

    return false unless exception.respond_to?(:backtrace)

    (Patterns::JAVA =~ exception.backtrace.first) != nil
  end

  def stack_frame(regexp, stackframe)
    if (match = match_frame(regexp, stackframe))
      return {
        in_app: sqreen_code?(match[:file]),
        abs_path: match[:file],
        function: match[:function],
        lineno: (Integer(match[:line]) if match[:line]),
      }
    end

    logger.error(
      "can't parse '#{stackframe}' (please file an issue so we can fix " \
          'it: https://github.com/sqreen/ruby-sdk/issues/new)',
    )
    {
      in_app: false,
      function: stackframe,
    }
  end

  def sqreen_code?(abs_path)
    # Ruby uses / on Windows too
    abs_path.include?('/lib/sqreen/')
  end

  def match_frame(regexp, stackframe)
    match = regexp.match(stackframe)
    return match if match

    Patterns::GENERIC.match(stackframe)
  end
end
